/*
 * Atom.cpp
 *
 * Copyright Notice: see copyright.txt
 *
 * Date: 3/8/2012 (D/M/Y)
 * Author: Dmitrij Lioubartsev
 * Email: dmitrijl42@gmail.com
 *
 * See Atom.h for more information about this class.
 */

#include "Atom.h"
#include <cmath>

Atom::Atom() {
	x = y = z = 0;
	element = "";
	distToSrc = DEFAULT_DIST;
	charge = 0;
	ff_type = "";
}

Atom::Atom(string name, double x, double y, double z) {
	element = name;
	setx(x);
	sety(y);
	setz(z);
	distToSrc = DEFAULT_DIST;
	charge = 0;
	ff_type = "";
}

Atom::~Atom() {
//do nothing
}

void Atom::setx(double xc) { x = xc;}
void Atom::sety(double yc) { y = yc;}
void Atom::setz(double zc) { z = zc;}
void Atom::setElement(string name) {element = name;}
void Atom::setDistToSrc(int dist) { distToSrc = dist;}
void Atom::setCharge(double charge) {this->charge = charge;}
void Atom::setFF_type(string type) {ff_type = type;}

double Atom::getx() const {return x;}
double Atom::gety() const {return y;}
double Atom::getz() const {return z;}
string Atom::getElement() {return element;}
int Atom::getDistToSrc() const {return distToSrc;}
double Atom::getCharge() {return charge;}
string Atom::getFF_type() {return ff_type;}

/*
 * Overloading == operator to compare atoms.
 * Compares their coordinates and element.
 */
int Atom::operator==(Atom& a) const {

	//This comparision is to prevent rounding errors.
	double xd = abs(x-a.getx());
	double yd = abs(y-a.gety());
	double zd = abs(z-a.getz());

	return ((a.getElement() == element) &&
			(xd < 10e-3 || xd > 0.999) &&
			(yd < 10e-3 || yd > 0.999) &&
			(zd < 10e-3 || zd > 0.999));
}

/*
 * This uses toCubic() on copies of both atoms before using == to compare.
 */
bool Atom::equals(Atom& a) {
	//copy atoms and make them "small" to
	//account for structure symmetry.
	Atom thiscpy(element,x,y,z);
	thiscpy.toCubic();
	Atom thatcpy(a.getElement(),a.getx(),a.gety(),a.getz());
	thatcpy.toCubic();

	return thiscpy == thatcpy;
}

/*
 * This takes the current x,y,z coordinates and sets
 * each of them to the relative place in the interval
 * [0,var[, where var is a,b and respectively c.
 */
void Atom::toCubic(double a, double b, double c) {
	//basically, only keep the decimal parts of the double,
	//(sets it to value between [0,1[), and then multiply by
	//the given parameter.
	if(x >= 0) {
		x = x - std::floor(x);
	} else { //x < 0
		x = std::abs(x);
		x = 1 - (x - std::floor(x));
	}
	x *= a;

	if(y >= 0) {
		y = y - std::floor(y);
	} else { //y < 0
		y = std::abs(y);
		y = 1 - (y - std::floor(y));
	}
	y*= b;

	if(z >= 0) {
		z = z - std::floor(z);
	} else { //z < 0
		z = std::abs(z);
		z = 1 - (z - std::floor(z));
	}
	z*= c;
}

/*
 * Check the neighbors of this atom, and return true
 * if it matches the Exclusion.
 */
bool Atom::checkPattern(Exclusion exs) {
	bool pattern = false;

	for(unsigned int i = 0; i < bonds.size(); i++) { //check each binded atom
		if(exs.getBinder() == DOESNT_MATTER) { //if bounding atom doesn't matter
			if((element == exs.getFirst() && (*(bonds[i])).hasBond(exs.getSecond(),this))
					|| (element == exs.getSecond() && (*(bonds[i])).hasBond(exs.getFirst(),this))) {
				//check both ways (Eg. Exclusion Al-O-Br checks Al-O-Br and Br-O-Al)
				pattern = true;
				break;
			}
		} else { //if it does matter, then check it as well.
			if((*(bonds[i])).getElement() == exs.getBinder()) //add this extra check
			if((element == exs.getFirst() && (*(bonds[i])).hasBond(exs.getSecond(),this))
					|| (element == exs.getSecond() && (*(bonds[i])).hasBond(exs.getFirst(),this))) {
				pattern = true;
				break;
			}
		}
	}

	return pattern;
}

/*
 * Check if the atom has a bond to an atom of type el,
 * but ignoring to check the Atom given in notThis.
 */
bool Atom::hasBond(string el, Atom* notThis) {
	bool hasBond = false;
	for(unsigned int i = 0;i < bonds.size();i++) {
		//if((*(bonds[i])).getElement() == el && !((*(bonds[i])) == (*notThis))) {
		if((*(bonds[i])).getElement() == el && !((bonds[i]) == notThis)) { //compare the pointers
			hasBond = true;
			break;
		}
	}
	return hasBond;
}


