/*
 * ForceField.cpp
 *
 * Copyright Notice: see copyright.txt
 *
 *  Date: 3/8/2012 (D/M/Y)
 *  Author: Dmitrij Lioubartsev
 *  Email: dmitrijl42@gmail.com
 *
 *  For class info, see ForceField.h
 */

#include "ForceField.h"

/*
 * allAtoms is used to determine which atoms that are relevant to read
 * from the force field file. (eg., there are no C atoms in allAtoms,
 * so all info about C in the force field file will be ignored.)
 */
ForceField::ForceField(vector<Atom>* allAtoms) {
	vector<Atom>& atoms = (*allAtoms);

	 //initialize atomTypes.
	for(unsigned int i = 0; i < atoms.size(); i++) {
		atomTypes.insert(atoms[i].getElement());
	}

	//initialize ff
	set<string>::iterator it;
	for(it = atomTypes.begin(); it != atomTypes.end(); it++) {
		vector<ForceFieldJoint> tmp;
		ff.insert(pair<string, vector<ForceFieldJoint> >(*it,tmp)); //initialize the element with an empty vector.
	}
}

ForceField::~ForceField() {
	// do nothing
}


/*
 * Load the information from the force field file.
 * If the file doesn't exist, returns false (else returns true)
 */
bool ForceField::loadFromFile(string ff_file) {
	ifstream ffFile(ff_file.c_str());
	if(!ffFile) {
		ffFile.close();
		return false; //file doesn't exist
	}

	bool done = false;
	bool eof = false;
	bool beginExists = false;
	while(!done && !eof) { //find begin_definitions
		//skip whitespaces and words until "begin_definitions" is found
		if(skipWhiteSpace(ffFile)) { //skip whitespaces
			eof = true;
			break;
		}
		string word;
		if(getWord(ffFile, word)) { //get word
			eof = true;
			break;
		}
		if(word == "begin_definitions" || word == "BEGIN_DEFINITIONS") {
			done = true;
			beginExists = true;
		}
	}

	if(eof) {
		cout << "WARNING: found no relevant data in the Force Field file."
				" Either it does not exist or it is not within a \"begin_"
				"definitions\" tag." << endl;
		return true; //file exists, so return true anyway
	}

	done = false;
	while(!done && !eof) {
		if(skipWhiteSpace(ffFile)) { //skip whitespace to element
			eof = true;
			break;
		}

		string element("");
		if(getWord(ffFile, element)) {
			eof = true;
		}

		if(element == "end_definitions" || element == "END_DEFINITIONS") {
			done = true;
			break;
		}

		if(eof)
			break;

		if(skipWhiteSpace(ffFile)) {//skip whitespace to number
			eof = true;
			break;
		}

		//now get number of elements to follow
		string numAtomsStr("");
		if(getWord(ffFile, numAtomsStr)) {
			eof = true;
			break;
		}

		int numAtoms = atoi(numAtomsStr.c_str());
		//now get the list of elements that define the bonds
		vector<string> atoms;
		if(skipWhiteSpace(ffFile)) {
			eof = true;
			break;
		}

		for(int i = 0; i < numAtoms; i++) {
			string tmp("");
			if(getWord(ffFile, tmp)) {
				eof = true;
				break;
			}
			if(skipWhiteSpace(ffFile)) {
				eof = true;
				break;
			}

			atoms.push_back(tmp);
		}
		if(ffFile.eof() || eof) //if suddenly eof in middle of that inner loop
			break;
		//now extract data, first the charge
		string dataDoubleStr("");
		if(getWord(ffFile, dataDoubleStr)) {
			eof = true;
			break;
		}
		double charge = atof(dataDoubleStr.c_str());
		if(skipWhiteSpace(ffFile)) {
			eof = true;
			break;
		}
		//then the force field type
		string ff_type("");
		if(getWord(ffFile,ff_type))
			eof = true;

		//create data types
		ForceFieldJoint::data dat = {charge, ff_type};
		ForceFieldJoint ffj(element, &atoms, dat);
		//and insert into the Force Field.
		if(atomTypes.find(element) != atomTypes.end())
			(ff[element]).push_back(ffj);
	}

	if(!done && beginExists) {
		cout << "WARNING: found no \"end_definitions\" in Force Field"
				" file. Any data about charge and type was still read"
				" but it is recommended to check the file." << endl;
	}

	return true;
}

/*
 * This is the old function, it assumes there are only this relevant data
 * in the file. The new version skips until it finds begin_definitions, then
 * reads all data until it finds end_definitions or eof.
 */
/*
bool ForceField::loadFromFile(string ff_file) {
	//start with clearing
	//for(map<string, vector<ForceFieldJoint> >::iterator it = ff.begin(); it != ff.end(); it++) {
	//	it->second.clear();
	//}

	ifstream ffFile(ff_file.c_str());
	if(!ffFile) {
		ffFile.close();
		return false; //file doesn't exist
	}

	while(!ffFile.eof()) {
		if(skipWhiteSpace(ffFile))
			break;

		string element("");
		if(getWord(ffFile, element))
			break;
		if(element == "begin_definitions" || element == "end_definitions") {
			//do nothing, i dont really care for this.
		} else { //assuming its a force field joint then.
			if(skipWhiteSpace(ffFile)) //skip whitespace to number
				break;
			//now get number of elements to follow
			string numAtomsStr("");
			if(getWord(ffFile, numAtomsStr))
				break;
			int numAtoms = atoi(numAtomsStr.c_str());
			//now get the list of elements that define the bonds
			vector<string> atoms;
			if(skipWhiteSpace(ffFile))
				break;
			for(int i = 0; i < numAtoms; i++) {
				string tmp("");
				if(getWord(ffFile, tmp))
					break;
				if(skipWhiteSpace(ffFile))
					break;
				atoms.push_back(tmp);
			}
			if(ffFile.eof()) //if suddenly eof in middle of that inner loop
				break;
			//now extract data, first the charge
			string dataDoubleStr("");
			if(getWord(ffFile, dataDoubleStr))
				break;
			double charge = atof(dataDoubleStr.c_str());
			if(skipWhiteSpace(ffFile))
				break;
			//then the force field type
			string ff_type("");
			getWord(ffFile,ff_type);

			//create data types
			ForceFieldJoint::data dat = {charge, ff_type};
			ForceFieldJoint ffj(element, &atoms, dat);
			//and insert into the Force Field.
			if(atomTypes.find(element) != atomTypes.end())
				(ff[element]).push_back(ffj);
		}

	}

	ffFile.close();
	return true;
}
*/

/*
 * Equalizes all atoms' charges, so that the total becomes total_Q
 */
void ForceField::setCharge(double total_Q, vector<Atom>* atoms, bool equalize) {
	unsigned int numAtoms = (*atoms).size();
	double totCharge = 0;
	//first initialize all charges and types
	int errors = 0;
	for(unsigned int i = 0; i < numAtoms; i++) {
		ForceFieldJoint::data dat;
		try {
			dat = getData((*atoms)[i]);
		}
		catch(std::out_of_range& oor) {
			errors++;
			if(errors < 10) //this is to avoid too much spam
			cout << "WARNING: Could not find any force field values "
					"for the element " << (*atoms)[i].getElement() << "." << endl;
			else if(errors == 10){
				cout << "NOTE: There are more errors like the 10 above"
						"but they are not printed out, to not spam stdout." << endl;
			}
			continue;
		}
		((*atoms)[i]).setCharge(dat.charge);
		((*atoms)[i]).setFF_type(dat.fft);
		totCharge += dat.charge;
	}
	//now check if total charge equals the input.
	if(fabs(totCharge - total_Q) > 0.0001 && equalize) {

		cout << "Total charge before equalizing operation: " << totCharge << endl;

		double diff = (total_Q - totCharge)/numAtoms; //the difference

		//to all atoms, add diff to their charge.
		for(unsigned int i = 0; i < numAtoms; i++) {
			((*atoms)[i]).setCharge(((*atoms)[i]).getCharge() + diff);
		}
	} else {
		cout << "Total charge: " << totCharge << ". No equalizing operations executed." << endl;
	}
}

/*
 * Return the data of the atom a.
 * @throws std::out_of_range If the element of 'a' does not exist
 * in the force field file.
 */
ForceFieldJoint::data ForceField::getData(Atom& a) {
	vector<ForceFieldJoint>& ffjVector = ff.at(a.getElement());
	int useThisFFj = -1;
	for(unsigned int i = 0; i <ffjVector.size();i++) {
		ForceFieldJoint& ffj = ffjVector[i];
		if(ffj.getBondCount() == 0) {
			if(useThisFFj == -1) {
				useThisFFj = i;
			}
		} else {
			if(ffj.matches(a.getBonds())) {
				useThisFFj = i;
			}
		}
	}

	if(useThisFFj == -1) { //if did not find data
		ForceFieldJoint::data dat = {0, "NULL"};
		return dat;
	}

	return ffjVector[useThisFFj].getData();

}

bool ForceField::skipWhiteSpace(ifstream& stream) {
	char c = ' ';

	bool eof = false;
	while(isWhiteSpace(c)) {
		stream.get(c);
		if(stream.eof()) {
			eof = true;
			break; //reached end of file, stop
		}
		if(c == '#')  {//comment
			stream.ignore(2048, '\n');
			if(stream.eof()) {
				eof = true;
				break;
			}
		}
	}
	stream.unget();
	return eof;
}

bool ForceField::getWord(ifstream& stream, string& word) {
	word = "";
	char c;
	bool eof = false;

	while(!(isWhiteSpace(c))) {
		stream.get(c);
		if(stream.eof()) {
			eof = true;
			break; //reached end of file, stop
		}
		word.append(1,c);
	}
	if(!eof) {
		stream.unget();
		word.erase(word.length()-1,1);
	}
	return eof;
}

bool ForceField::isWhiteSpace(char c) {
	return (c == '\n' || c == '\t' || c == ' ' || c == '#' || c == '\r');
}
